import { useDisclosure } from "@mantine/hooks";
import cx from "classnames";
import { useState } from "react";

import type { SdkUsageProblem } from "embedding-sdk/types/usage-problem";
import LogoIcon from "metabase/components/LogoIcon";
import ExternalLink from "metabase/core/components/ExternalLink";
import { originalColors } from "metabase/lib/colors";
import {
  Button,
  Card,
  Flex,
  Icon,
  Popover,
  Stack,
  Text,
  useMantineTheme,
} from "metabase/ui";

import S from "./SdkUsageProblemBanner.module.css";

export interface SdkUsageProblemBannerProps {
  problem: SdkUsageProblem | null;
}

// Prevent the usage problem banner from inheriting the theme colors,
// so they remain legible even when the theme is changed.
const unthemedBrand = originalColors["brand"];
const unthemedTextDark = originalColors["text-dark"];
const unthemedTextMedium = originalColors["text-medium"];

export const SdkUsageProblemBanner = ({
  problem,
}: SdkUsageProblemBannerProps) => {
  const theme = useMantineTheme();

  const [expanded, setExpanded] = useState(false);
  const [shown, { close: hideBanner }] = useDisclosure(true);

  if (!problem || !shown) {
    return null;
  }

  const { severity } = problem;
  const isError = severity === "error";

  // When the font family cannot be loaded from the MB instance,
  // due to MB instance outage or missing CORS header,
  // we fall back to the system font.
  const fontFamily = `${theme.fontFamily}, sans-serif`;

  return (
    <Popover
      position="top-start"
      opened={expanded}
      onChange={setExpanded}
      // We don't want to render the popover within the portal,
      // as this is used within a fixed-position container in a portal.
      withinPortal={false}
    >
      <Popover.Target>
        <Flex
          onClick={() => setExpanded(!expanded)}
          className={cx(S.Indicator, isError ? S.Error : S.Warning)}
          data-testid="sdk-usage-problem-indicator"
          i-should-be-flex="true"
        >
          <Flex bg="white" px="sm" py="xs" className={S.Logo}>
            <LogoIcon height={24} fill={unthemedBrand} />
          </Flex>

          <Flex
            justify="center"
            align="center"
            px="9px"
            columnGap="sm"
            className={S.Content}
          >
            <Icon
              name={isError ? "warning_round_filled" : "warning"}
              size={14}
              fill={unthemedTextDark}
              stroke={isError ? unthemedTextDark : undefined}
            />

            <Text tt="capitalize" c={unthemedTextMedium} ff={fontFamily}>
              {severity}
            </Text>
          </Flex>
        </Flex>
      </Popover.Target>

      <Popover.Dropdown className={S.PopoverDropdown}>
        <Card
          p="md"
          radius="md"
          maw="20rem"
          data-testid="sdk-usage-problem-card"
        >
          <Stack gap="sm">
            <Flex w="100%" justify="space-between">
              <Text
                fw="bold"
                size="lg"
                tt="capitalize"
                c={unthemedTextDark}
                ff={fontFamily}
              >
                {severity}
              </Text>

              <Icon
                name="chevrondown"
                size={16}
                onClick={() => setExpanded(false)}
                cursor="pointer"
                fill={unthemedTextDark}
              />
            </Flex>

            <Text c={unthemedTextDark} ff={fontFamily}>
              {problem.message}
            </Text>

            <Flex w="100%" justify="end" mt="sm" columnGap="sm">
              <Button
                fz="sm"
                variant="subtle"
                onClick={hideBanner}
                ff={fontFamily}
                size="compact-md"
              >
                Hide {severity}
              </Button>

              <ExternalLink role="link" href={problem.documentationUrl}>
                <Button
                  fz="sm"
                  variant="outline"
                  rightSection={<Icon name="external" size={10} />}
                  ff={fontFamily}
                  className={S.DocsButton}
                  size="compact-md"
                >
                  View documentation
                </Button>
              </ExternalLink>
            </Flex>
          </Stack>
        </Card>
      </Popover.Dropdown>
    </Popover>
  );
};
